import dataclasses
import datetime
import os
import pathlib
import re
import typing
import urllib.parse

from . import commands
from . import githost


TODAY = datetime.date.today()
SSH_GIT_URL = re.compile('(?P<user>.+)@(?P<host>.+):(?P<path>.+)')
DEFAULT_API_URL = "https://api.github.com"


@dataclasses.dataclass
class GitHostApiAccessInfo:
    api_base_url: str = DEFAULT_API_URL
    auth_token_file: str = None
    auth_token_env_var: str = None

    def __post_init__(self):
        self.auth_token = self.get_auth_token()

    def get_auth_token(self):
        if self.auth_token_file:
            return pathlib.Path(self.auth_token_file).expanduser().read_text(encoding='utf-8').strip()
        if self.auth_token_env_var:
            return os.environ[self.auth_token_env_var]
        raise ValueError(
            "Missing configuration to get authentication token for Git host API."
        )


@dataclasses.dataclass
class Config:
    platform: str = None
    path: str = "."
    max_age: int = 90

    colorize_errors: bool = True
    xunit_file: str = None

    calm_branches: typing.Sequence = ("gh-pages", "master", "prod", "maint(enance)?/.*")
    ignore_branches_without_pull_request: bool = False

    host_api_access: dict = None

    # Cannot be configured, but is automatically filled in instead. Default value is irrelevant and only here to
    # initialize the field.
    host: str = dataclasses.field(default=None, init=False)
    host_owner: str = dataclasses.field(default=None, init=False)
    repo_name: str = dataclasses.field(default=None, init=False)

    def __post_init__(self):
        if self.host_api_access:
            self.host_api_access = GitHostApiAccessInfo(**self.host_api_access)

        self.host, self.host_owner, self.repo_name = get_repository_info(self.path)

        if not self.platform:
            if "gitlab" in self.host.lower():
                self.platform = "gitlab"
                if self.host_api_access and self.host_api_access.api_base_url == DEFAULT_API_URL:
                    self.host_api_access.api_base_url = f"https://{self.host}/api/v4"
            else:
                self.platform = "github"  # non-breaking change: keep the previous default value

    def ignore_branch(self, branch):
        for calm_branch_regexp in self.calm_branches:
            if re.match(calm_branch_regexp, branch):
                return True
        return False

    def get_branch_url(self, branch):
        return f"https://{self.host}/{self.host_owner}/{self.repo_name}/tree/{branch}"


@dataclasses.dataclass
class BranchInfo:
    repo: str
    name: str
    url: str
    author: str
    age: int
    is_old: bool
    pull_request: githost.PullRequestInfo = None

    @property
    def name_and_details(self):
        details = f"{self.name} ({self.url})"
        if self.pull_request:
            pr = self.pull_request
            details += f", linked to {pr.state} PR/MR #{pr.number} ({pr.url})"
        return details


def get_repository_info(path):
    """Extract the repository owner and name from the origin remote."""
    # Compared to "git config --get remote.origin.url" or "git remote get-url origin", this handles aliases defined in
    # git config.
    remote_url = commands.get_output(("git", "ls-remote", "--get-url"), cwd=path)[0]
    # It usually looks like "git@github.com:Polyconseil/check-oldies.git"
    # but could also look like "poly:check-fixmes.git" if you have an
    # "[url]" section with "insteadOf" in your Git configuration.
    if remote_url.endswith(".git"):
        remote_url = remote_url[:-4]
    parsed = urllib.parse.urlparse(remote_url)
    if parsed.scheme in ('http', 'https', 'ssh'):
        path = parsed.path.lstrip('/')
        host_owner, repo_name = path.split('/', 1)
        # Drop "git@" prefix with ssh scheme
        host = parsed.netloc.split('@')[-1] if parsed.scheme == "ssh" else parsed.netloc
        return host, host_owner, repo_name
    ssh_match = SSH_GIT_URL.match(remote_url)
    if ssh_match:
        path = ssh_match.group('path')
        host_owner, repo_name = path.split('/', 1)
        return ssh_match.group('host'), host_owner, repo_name
    raise ValueError(f"Could not parse remote origin and determine the Git host: '{remote_url}'")


def get_branches(config: Config):
    branches = []
    for branch in commands.get_output(("git", "branch", "--remotes"), cwd=config.path):
        branch = branch.strip()
        if not branch.startswith("origin/"):
            continue
        branch = branch.strip()[len("origin/") :]
        if config.ignore_branch(branch) or "->" in branch:
            continue
        output = commands.get_output(
            ("git", "log", f"origin/{branch}", "-1", "--format=%ae %ci"),
            cwd=config.path,
        )[0]
        # line looks like "john.smith@mail.test 2018-12-19 14:18:52 +0100"
        email, date, *_rest = output.split(" ")
        date = datetime.date(*[int(s) for s in date.split("-")])
        age = (TODAY - date).days
        branches.append(
            BranchInfo(
                repo=config.repo_name,
                name=branch,
                url=config.get_branch_url(branch=branch),
                author=email,
                age=age,
                is_old=age > config.max_age,
            )
        )

    if not branches:
        return ()

    if config.host_api_access:
        pr_getter = githost.PullRequestGetter(config.platform, config.host_owner, config.host_api_access)
        for branch in branches:
            branch.pull_request = pr_getter.get_pull_request(config.repo_name, branch.name)

    return [
        branch
        for branch in branches
        if not config.ignore_branches_without_pull_request or branch.pull_request
    ]



import argparse
import os
import sys

from . import branches
from . import configuration
from . import xunit


def get_parser():
    parser = argparse.ArgumentParser(
        prog="check-branches", description="Check your code for unattended branches"
    )
    parser.add_argument(
        "--conf",
        help=(
            f"Path of the configuration file. "
            f"Defaults to {configuration.PYPROJECT_FILENAME} if it exists."
        ),
    )
    parser.add_argument(
        "path",
        nargs="?",
        help=(
            "Git-managed path where search should happen. "
            "Defaults to the working directory."
        ),
    )
    parser.add_argument(
        "--max-age",
        type=int,
        help=(
            f"Maximum age in days allowed for a branch, errors otherwise. "
            f"Defaults to {branches.Config.max_age}."
        ),
    )
    parser.add_argument(
        "--no-color",
        action="store_false",
        default=True,
        dest="colorize_errors",
        help="Do not colorize errors. Defaults to colorizing errors in red.",
    )
    parser.add_argument(
        "--xunit-file",
        action="store",
        help="Path of the xUnit report file to write. Defaults to no xUnit output.",
    )
    return parser


def branch_str(branch):
    return (
        f"{branch.author[:30]: <30} - {branch.age: >4} days - {branch.name_and_details}"
    )


def main():
    parser = get_parser()
    config = configuration.get_config(
        "check-branches", parser, sys.argv[1:], branches.Config
    )
    if not configuration.is_git_directory(config.path):
        sys.exit(f'Invalid path: "{config.path}" is not a Git repository.')

    if config.colorize_errors:
        warn = "\033[91m{}\033[0m".format
    else:
        warn = lambda text: text

    all_branches = branches.get_branches(config)

    out = []
    uncolorized_out = []
    for branch in sorted(
        all_branches, key=lambda branch: (branch.author, -branch.age, branch.name)
    ):
        line = branch_str(branch)
        out.append(warn(line) if branch.is_old else line)
        uncolorized_out.append(line if branch.is_old else line)
    has_old_branches = any(branch for branch in all_branches if branch.is_old)

    out = os.linesep.join(out)
    if has_old_branches:
        err_msg = "NOK: Some branches are too old."
        print(err_msg)
    else:
        err_msg = ""
        print("OK: All branches are fresh.")
    if out:
        print(out)

    if config.xunit_file:
        uncolorized_out = os.linesep.join(uncolorized_out)
        xunit.create_xunit_file(
            os.path.abspath(config.xunit_file),
            "check-branches",
            "branches",
            "CheckBranches",
            err_msg=err_msg,
            stdout=uncolorized_out,
            stderr="",
        )

    sys.exit(os.EX_DATAERR if has_old_branches else os.EX_OK)


if __name__ == "__main__":  # pragma: no cover
    main()



import os
from xml.etree import ElementTree


def create_xunit_file(
    output_path,
    suite_name,
    case_name,
    class_name,
    err_msg,
    stdout,
    stderr,
):
    """Create an xUnit file.

    If the parent directory of ``output_path`` does not exist, it is
    transparently created.
    """
    suite = ElementTree.Element(
        "testsuite",
        {
            "name": suite_name,
            "tests": "1",
            "errors": "0",
            "failures": "1" if err_msg else "0",
        },
    )

    case = ElementTree.SubElement(
        suite, "testcase", {"classname": class_name, "name": case_name}
    )
    if err_msg:
        ElementTree.SubElement(case, "failure", {"message": err_msg})
    system_out = ElementTree.SubElement(case, "system-out")
    system_out.text = stdout

    system_err = ElementTree.SubElement(case, "system-err")
    system_err.text = stderr

    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    ElementTree.ElementTree(suite).write(output_path, encoding="utf-8")



import os
import pipes
import subprocess
import tempfile


def get_output(cmd_list, cwd, valid_return_codes=(0,)):
    """Return command output as a list of non-empty lines."""
    res = subprocess.run(
        cmd_list,
        cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        encoding="utf-8",
    )
    if res.returncode not in valid_return_codes:
        raise subprocess.CalledProcessError(
            res.returncode, cmd_list, res.stdout, res.stderr,
        )
    # Remove empty lines.
    return list(filter(bool, res.stdout.split(os.linesep)))


def get_pipe_command_output(base_cmd_list, piped_to, cwd=None, valid_return_codes=(0,)):
    """Return output of ``base_cmd_list | piped_to`` as a list of
    non-empty lines.
    """
    _fp, pipefile_path = tempfile.mkstemp()
    try:
        pipe = pipes.Template()
        pipe.append(piped_to, "--")
        pipefile = pipe.open(pipefile_path, "w")
        command_output = get_output(
            base_cmd_list, cwd=cwd, valid_return_codes=valid_return_codes
        )
        pipefile.write(os.linesep.join(command_output))
        pipefile.close()
        with open(pipefile_path, encoding='utf-8') as fp:
            lines = fp.read().split(os.linesep)
        return list(filter(bool, lines))
    finally:
        if not pipefile.closed:  # pragma: no cover
            pipefile.close()
        os.unlink(pipefile_path)



import argparse
import collections
import dataclasses
import datetime
from email.message import EmailMessage
import pathlib
import smtplib
import sys
import typing

from . import annotations
from . import branches
from . import configuration


CATCH_ALL = object()
CONFIGURATION_FILE = 'forget-me-not.toml'


@dataclasses.dataclass
class RecipientConfig:
    catch_all: str
    list: typing.Sequence
    aliases: dict = dataclasses.field(default_factory=dict)

    @property
    def email_mapping(self):
        if hasattr(self, '_mapping'):
            return self._mapping  # pylint: disable=access-member-before-definition
        self._mapping = {}
        # john.smith@example.com gets two aliases: "john.smith" and "jsmith".
        for email in self.list:
            fullname = email.split('@')[0]
            self._mapping[fullname] = email
            if '.' in fullname:
                firstname, lastname = fullname.split('.')
                # XXX: duplicate aliases are not handled
                self._mapping[f'{firstname[0]}{lastname}'] = email
        self._mapping.update(self.aliases)
        return self._mapping

    def get_email(self, author):
        if '@' in author:  # assume it's an e-mail address
            return author
        return self.email_mapping.get(author)


@dataclasses.dataclass
class SmtpConfig:
    host: str
    port: int = 0  # use OS default behaviour

    def make_connection(self):
        return smtplib.SMTP(self.host, port=self.port)


@dataclasses.dataclass
class Config:
    recipients: dict
    path: str = '.'
    warning_delay: int = 15
    ignored_repositories: typing.Sequence = ()
    output: typing.Sequence = ("stdout", )

    smtp: dict = dataclasses.field(default_factory=lambda: {'host': 'localhost'})

    email_sender: str = "forget-me-not@example.com"
    email_subject: str = "W{week_number}: Old annotations and old branches"
    email_body_intro: str = "    Hello,\n\n"
    email_body_outro: str = "\n\n-- \nSent by forget-me-not"
    email_body_annotations: str = "\n".join((
        "You have {count} old annotation(s) that will block builds "
        "in less than {warning_delay} days:",
        "",
        "{lines}",
    ))
    email_body_annotation_line_template: str = (
        "{age} days: {repo}:{path}:{line_no}: {line_content}"
    )
    email_body_branches: str = "\n".join((
        "You have {count} old branch(es) that will block builds "
        "in less than {warning_delay} days:",
        "",
        "{lines}",
    ))
    email_body_branch_line_template: str = "{age} days: {repo}: {branch_details}"


    def __post_init__(self):
        assert set(self.output).issubset({"stdout", "mail"})
        self.recipients = RecipientConfig(**self.recipients)
        if self.smtp:
            self.smtp = SmtpConfig(**self.smtp)


@dataclasses.dataclass
class FakeArgumentParser:
    # An object that mimics what we get from ArgumentParser.parse_args().
    path: pathlib.Path
    conf: pathlib.Path

    def parse_args(self, argv):
        return self


def check_repositories(config):
    all_reports = {
        'annotations': [],
        'branches': []
    }
    for path in sorted(pathlib.Path(config.path).iterdir()):
        if not path.is_dir():
            continue
        if path.stem in config.ignored_repositories:
            continue
        if not configuration.is_git_directory(path):
            continue
        repo_reports = check_repository(path, config.warning_delay)
        all_reports["annotations"].extend(repo_reports["annotations"])
        all_reports["branches"].extend(repo_reports["branches"])
    return all_reports


def check_repository(path, warning_delay):
    repo_config_path = path / configuration.PYPROJECT_FILENAME
    if not repo_config_path.exists():
        repo_config_path = None  # we'll use the default config
    ann_config = configuration.get_config(
        tool_name="check-fixmes",
        arg_parser=FakeArgumentParser(path=path, conf=repo_config_path),
        argv=[],
        config_class=annotations.Config,
    )
    branches_config = configuration.get_config(
        tool_name="check-branches",
        arg_parser=FakeArgumentParser(path=path, conf=repo_config_path),
        argv=[],
        config_class=branches.Config,
    )
    ann_config.max_age -= warning_delay
    branches_config.max_age -= warning_delay
    all_annotations = annotations.get_annotations(ann_config)
    for annotation in all_annotations:
        annotation.repository = path.stem
    all_branches = branches.get_branches(branches_config)
    return {
        "annotations": [ann for ann in all_annotations if ann.is_old],
        "branches": [branch for branch in all_branches if branch.is_old],
    }


def group_reports_by_email(reports, config):
    by_email = collections.defaultdict(dict)
    unknown_users = set()

    for report_key, recipient_getter in (
        ("annotations", lambda ann: ann.assignee),
        ("branches", lambda branch: branch.author),
    ):
        for warning in reports[report_key]:
            recipient = recipient_getter(warning)
            email = config.recipients.get_email(recipient)
            if not email:
                unknown_users.add(recipient)
                email = CATCH_ALL
            by_email[email].setdefault(report_key, []).append(warning)

    return by_email, unknown_users


def generate_emails(reports_by_email, config):
    week_number = datetime.date.today().isocalendar()[1]
    email_subject = config.email_subject.format(week_number=week_number)
    for recipient, reports in reports_by_email.items():
        body = config.email_body_intro
        all_annotations = reports.get("annotations")
        if all_annotations:
            # If no user is assigned, promote oldest lines because
            # they are the most urgent to fix. Otherwise, keep the
            # original order (by repository, filename and line number)
            # because it makes more sense to group annotations in each
            # repository.
            if recipient == CATCH_ALL:
                all_annotations.sort(key=lambda ann: ann.age)
            lines = []
            for annotation in all_annotations:
                lines.append(
                    config.email_body_annotation_line_template.format(
                        age=annotation.age,
                        repo=annotation.repository,
                        path=annotation.filename,
                        line_no=annotation.line_no,
                        line_content=annotation.line_content.strip(),
                    )
                )
            body += config.email_body_annotations.format(
                count=len(lines),
                warning_delay=config.warning_delay,
                lines="\n".join(lines)
            )

        all_branches = reports.get("branches")
        if all_branches:
            # Sort as we did for annotations above.
            if recipient == CATCH_ALL:
                all_branches.sort(key=lambda ann: ann.age)
            lines = [
                config.email_body_branch_line_template.format(
                    age=branch.age,
                    repo=branch.repo,
                    branch_details=branch.name_and_details,
                )
                for branch in all_branches
            ]
            if all_annotations:
                body += '\n\n'
            body += config.email_body_branches.format(
                count=len(lines),
                warning_delay=config.warning_delay,
                lines='\n'.join(lines),
            )

        body += config.email_body_outro

        msg = EmailMessage()
        msg.set_content(body)
        msg["Subject"] = email_subject
        msg["From"] = config.email_sender
        if recipient == CATCH_ALL:
            msg["To"] = config.recipients.catch_all
        else:
            msg["To"] = recipient
        yield msg


def get_parser():
    parser = argparse.ArgumentParser(
        prog="forget-me-not",
        description="Warn about annotations and branches that will soo be too old."
    )
    parser.add_argument(
        'path',
        nargs='?',
        help=(
            "Directory that contains all Git repositories to check. "
            "Defaults to the working directory."
        ),
    )
    parser.add_argument(
        "--conf",
        help=(
            f"Path of the configuration file. "
            f"Defaults to {CONFIGURATION_FILE} if it exists."
        ),
    )
    parser.add_argument(
        "--warning-delay",
        type=int,
        help=(
            f"Number of days. "
            f"Annotations and branches are reported if they turn old within this delay. "
            f"Defaults to {Config.warning_delay}."
        ),
    )
    parser.add_argument(
        '--output', choices=['mail', 'stdout'], action='append', default=['stdout'],
    )
    return parser


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = get_parser()
    config = configuration.get_config(
        None,
        parser,
        argv,
        Config,
        'forget-me-not.toml',
    )

    reports = check_repositories(config)
    reports_by_email, unknown_users = group_reports_by_email(reports, config)
    emails = list(generate_emails(reports_by_email, config))

    if unknown_users:
        catch_all = config.recipients.catch_all
        print(
            f"Found annotations or old branches of unknown authors "
            f"(forwarded to {catch_all}):"
        )
        print('\n'.join(sorted(unknown_users)))

    if "stdout" in config.output:
        for email in emails:
            print(email["To"])
            print(email.get_content())
            print("-" * 20)
    if "mail" in config.output:
        with config.smtp.make_connection() as smtp:
            for email in emails:
                smtp.send_message(email)
                print(f"Sent e-mail to {email['To']}")


if __name__ == "__main__":  # pragma: no cover
    main()



import argparse
import os
import sys

from . import annotations
from . import configuration
from . import xunit


def annotation_str(annotation):
    return (
        f"{annotation.assignee: <15} - {annotation.age: >4} days - "
        f"{annotation.filename}:{annotation.line_no}: {annotation.line_content.strip()}"
    )


def orphan_str(orphan):
    return (
        f"{orphan.author: <15} -   ORPHAN  - "
        f"{orphan.path}:{orphan.line_no}: Unknown tag {orphan.tag}"
    )


def get_parser():
    parser = argparse.ArgumentParser(
        prog="check-fixmes", description="Check your code for unattended annotations"
    )
    parser.add_argument(
        "--conf",
        help=(
            f"Path of the configuration file. "
            f"Defaults to {configuration.PYPROJECT_FILENAME} if it exists."
        ),
    )
    parser.add_argument(
        "path",
        nargs="?",
        help=(
            "Git-managed path where search should happen. "
            "Defaults to the working directory."
        ),
    )
    parser.add_argument(
        "--max-age",
        type=int,
        help=(
            f"Maximum age in days allowed for an annotation, errors otherwise. "
            f"Defaults to {annotations.Config.max_age}."
        ),
    )
    parser.add_argument(
        "--no-color",
        action="store_false",
        default=True,
        dest="colorize_errors",
        help="Do not colorize errors. Defaults to colorizing errors in red.",
    )
    parser.add_argument(
        "--xunit-file",
        action="store",
        help="Path of the xUnit report file to write. Defaults to no xUnit output.",
    )
    return parser


def main():
    parser = get_parser()
    config = configuration.get_config(
        "check-fixmes", parser, sys.argv[1:], annotations.Config
    )
    if not configuration.is_git_directory(config.path):
        sys.exit(f'Invalid path: "{config.path}" is not a Git repository.')

    if config.colorize_errors:
        warn = "\033[91m{}\033[0m".format
    else:
        warn = lambda text: text

    # Look for old annotations
    out = []
    uncolorized_out = []
    all_annotations = sorted(
        annotations.get_annotations(config),
        key=lambda f: (f.assignee, -f.age, f.filename, f.line_no),
    )
    for annotation in all_annotations:
        line = annotation_str(annotation)
        out.append(warn(line) if annotation.is_old else line)
        uncolorized_out.append(line if annotation.is_old else line)
    has_old_annotations = any(ann for ann in all_annotations if ann.is_old)

    # Look for orphan FUTURE tags
    orphan_futures = annotations.get_orphan_futures(config)
    for orphan in orphan_futures:
        out.append(warn(orphan_str(orphan)))
        uncolorized_out.append(orphan_str(orphan))

    out = os.linesep.join(out)
    if has_old_annotations or orphan_futures:
        err_msg = "NOK: Some annotations are too old, or there are orphan FUTURE tags."
        print(err_msg)
    else:
        err_msg = ""
        if all_annotations:
            print("OK: All annotations are fresh.")
        else:
            print("OK: No annotations were found.")
    if out:
        print(out)

    if config.xunit_file:
        uncolorized_out = os.linesep.join(uncolorized_out)
        xunit.create_xunit_file(
            os.path.abspath(config.xunit_file),
            suite_name="check-fixmes",
            case_name="fixmes",
            class_name="CheckFixmes",
            err_msg=err_msg,
            stdout=uncolorized_out,
            stderr="",
        )

    sys.exit(os.EX_DATAERR if has_old_annotations or orphan_futures else os.EX_OK)


if __name__ == "__main__":  # pragma: no cover
    main()



import dataclasses
import json
import urllib.parse
import urllib.request


@dataclasses.dataclass
class PullRequestInfo:
    number: str
    state: str
    url: str


def PullRequestGetter(platform, host_owner, api_access):
    if platform.lower() == "github":
        klass = GitHubApi
    elif platform.lower() == "gitlab":
        klass = GitLabApi
    else:
        raise ValueError("Unsupported Git host platform." "")
    return klass(host_owner, api_access)


class GitHubApi:
    authentication_header = 'token'

    def __init__(self, host_owner, api_access):
        self.api_base_url = api_access.api_base_url
        self.owner = host_owner
        self.auth_token = api_access.auth_token

    def get_pull_request(self, repo, branch):
        url = f"{self.api_base_url}/repos/{self.owner}/{repo}/pulls"
        # If the branch is linked to more than one pull request,
        # return the most recent one.
        query = {
            "head": f"{self.owner}:{branch}",
            "state": "all",
            "sort": "created",
            "direction": "desc",
        }
        response = self._get(url, query)
        if not response:
            return None
        info = response[0]
        return PullRequestInfo(
            number=info["number"],
            state=info["state"],
            url=info["html_url"],
        )

    def _get(self, url, query=None):
        if query:
            url += "?" + urllib.parse.urlencode(query)
        headers = {
            "Authorization": f"{self.authentication_header} {self.auth_token}",
            "Content-Type": "application/json",
        }
        request = urllib.request.Request(url, headers=headers)
        response = urllib.request.urlopen(request)  # pylint: disable=consider-using-with
        return json.loads(response.read())


class GitLabApi(GitHubApi):
    authentication_header = 'Bearer'

    def _get_project(self, repo):
        query = {
            "search": f"{repo}",
            "scope": "projects",
        }
        response = self._get(f'{self.api_base_url}/search/', query)
        for project in response:
            if project['name'] == repo:
                return project

        return None

    def get_pull_request(self, repo, branch):
        project = self._get_project(repo)
        if not project:
            return None
        url = f"{self.api_base_url}/projects/{project['id']}/merge_requests"
        # If the branch is linked to more than one pull request,
        # return the most recent one.
        query = {
            "source_branch": f"{branch}",
            "order_by": "created_at",
            "sort": "desc",
        }
        response = self._get(url, query)
        if not response:
            return None
        info = response[0]
        return PullRequestInfo(
            number=info["iid"],
            state=info["state"],
            url=info["web_url"],
        )



import collections
import dataclasses
import datetime
import re
import subprocess
import typing

import pkg_resources

from . import commands


IGNORE_PRAGMA = "no-check-fixmes"


# All default values are here (not in the `config` module) to be close
# to the code that uses them.
@dataclasses.dataclass
class Config:
    path: str = "."
    max_age: int = 180

    colorize_errors: bool = True
    xunit_file: str = None

    annotations: typing.Sequence = ("todo", "fixme", )  # no-check-fixmes
    ignored_orphans_annotations: typing.Sequence = ("wontfix", "xxx")  # annotation which won't trigger orphans checks
    assignee_regex: str = r"(?:{annotation_regex})\s*\((?P<assignee>[\w\._-]+)"
    future_tag_regex: str = r"FUTURE-[-[:alnum:]\._]+?"  # no-check-fixmes

    whitelist: typing.Sequence = ()

    @property
    def annotation_regex(self):
        return "|".join(self.annotations).lower()

    @property
    def ignored_orphans_annotations_regex(self):
        return "|".join(self.ignored_orphans_annotations).lower()

    @property
    def _annotation_with_boundaries_regex(self):
        annotation_regex = r'\b|\b'.join(self.annotations).lower()
        return fr'\b{annotation_regex}\b'

    @property
    def py_annotation_regex(self):
        """Python compiled regex to extract an annotation."""
        # pylint: disable=access-member-before-definition
        if hasattr(self, "_py_annotation_regex"):
            return self._py_annotation_regex
        # Only accept tags if preceded by #, <, {, some space, /, * or line start
        self._py_annotation_regex = re.compile(
            fr"((?<=#|\s|/|\*|<|{{)|^)({self._annotation_with_boundaries_regex})",
            re.IGNORECASE | re.UNICODE,
        )
        return self._py_annotation_regex

    @property
    def py_assignee_regex(self):
        """Python compiled regex to extract the assignee."""
        # pylint: disable=access-member-before-definition
        if hasattr(self, "_py_assignee_regex"):
            return self._py_assignee_regex
        # Should look like "FIXME (jsmith)", the space is optional. [no-check-fixmes]
        self._py_assignee_regex = re.compile(
            self.assignee_regex.format(annotation_regex=self.annotation_regex),
            re.IGNORECASE | re.UNICODE,
        )
        return self._py_assignee_regex


@dataclasses.dataclass
class Annotation:
    filename: str
    line_no: int
    line_content: str
    age: int = None
    assignee: str = None
    is_old: bool = None


@dataclasses.dataclass
class FutureTag:
    path: str
    line_no: int
    tag: str
    author: str = None


def get_annotation_candidates(directory, annotation_regex, whitelist):
    """Return lines (with filename and line number) that contains an annotation."""
    return commands.get_output(
        [
            "git",
            "grep",
            "-I",  # Ignore binary files
            "--line-number",
            "--ignore-case",
            "--extended-regexp",
            "-e",  # Find annotations
            annotation_regex,
            "--and",  # When they do not contain the ignore pragma ("no-check-fixmes")
            "--not",
            "-e",
            IGNORE_PRAGMA,
            "--",
            ".",
        ]
        + [f":(exclude){glob}" for glob in whitelist],
        cwd=directory,
        valid_return_codes=(0, 1),  # 1 means that no files were found
    )


def get_line_blame(filename, line, cwd):
    """Return author's email and timestamp of the latest commit that
    touched this line.
    """
    infos = commands.get_output(
        ["git", "blame", f"-L {line},{line}", "--porcelain", "--", filename],
        cwd=cwd,
    )

    def _get_info(metadata):
        return [info.split()[1] for info in infos if info.startswith(metadata)][0]

    committer_mail = _get_info("committer-mail")
    timestamp = int(_get_info("committer-time"))
    timezone = _get_info("committer-tz")
    multiplier = 1 if timezone[0] == "+" else -1
    hours = int(timezone[1:3])
    minutes = int(timezone[3:5])
    tz = datetime.timezone(
        multiplier * datetime.timedelta(hours=hours, minutes=minutes)
    )
    commit_datetime = datetime.datetime.fromtimestamp(timestamp).astimezone(tz)
    return committer_mail, commit_datetime


def get_login_from_committer_email(committer_email):
    login_match = re.match(r"<(?P<login>.+)@", committer_email)
    if login_match:
        return login_match.group("login")
    return committer_email


def get_annotations(config: Config):
    annotations = []
    for candidate in get_annotation_candidates(
        config.path, config.annotation_regex, config.whitelist
    ):
        filename, line_no, line_content = candidate.split(":", 2)
        # FIXME (dbaty, 2020-10-21): it should be possible to apply
        # the right regex (with boundaries) directly in git grep.
        if config.py_annotation_regex.search(line_content):
            annotations.append(Annotation(filename, int(line_no), line_content))

    for annotation in annotations:
        last_committer, last_modification = get_line_blame(
            annotation.filename, annotation.line_no, cwd=config.path
        )
        last_committer = get_login_from_committer_email(last_committer)
        match = config.py_assignee_regex.search(annotation.line_content)
        annotation.assignee = match.group("assignee") if match else last_committer
        annotation.age = (
            datetime.datetime.now(datetime.timezone.utc) - last_modification
        ).days
        annotation.is_old = annotation.age > config.max_age
    return annotations


def get_known_future_tags(directory, annotation_regex, future_tag_regex, whitelist):
    """Return a list of tags that are referenced along annotations."""
    grep = [
        "git",
        "grep",
        "-I",  # ignore binary files
        "--ignore-case",
        "--extended-regexp",
        f"{annotation_regex}.*{future_tag_regex}",
        "--",
        ".",
    ]
    grep.extend([f":(exclude){glob}" for glob in whitelist])
    lines = commands.get_pipe_command_output(
        grep,
        piped_to=f'sed --regexp-extended "s/.*?({future_tag_regex}).*?/\\1/g"',
        cwd=directory,
        valid_return_codes=(0, 1),  # 0 if there are matches, 1 otherwise
    )
    return set(lines)


def git_supports_only_matching():
    out = subprocess.check_output(["git", "--version"]).decode("utf-8")
    # output looks like "git version 2.26.0"
    git_version = pkg_resources.parse_version(out.split()[2])
    # `git grep --only-matching` appeared in Git 2.19.0
    # https://github.com/git/git/blob/v2.19.0/Documentation/RelNotes/2.19.0.txt#L41
    minimal_version = pkg_resources.parse_version("2.19.0")
    return git_version >= minimal_version


def get_all_futures(directory, future_tag_regex, whitelist):
    """Get all occurrences of FUTURE tags."""
    # Old versions of Git (such as Git 2.1.4 that is shipped by Debian
    # Jessie) do not support the `--only-matching` option. Pipe to
    # `sed` instead.
    if git_supports_only_matching():
        grep = [
            "git",
            "grep",
            "-I",  # ignore binary files
            "--line-number",
            "--extended-regexp",
            "--only-matching",
            "-e",
            future_tag_regex,
            "--and",
            "--not",
            "-e",
            IGNORE_PRAGMA,
        ]
        grep.extend([f":(exclude){glob}" for glob in whitelist])
        lines = commands.get_output(
            grep,
            cwd=directory,
            valid_return_codes=(0, 1),  # 0 if there are matches, 1 otherwise
        )
    else:
        grep = [
            "git",
            "grep",
            "-I",  # ignore binary files
            "--line-number",
            "--extended-regexp",
            "-e",
            future_tag_regex,
            "--and",
            "--not",
            "-e",
            IGNORE_PRAGMA,
            "--",  # needed on old versions...
            ".",  # ... of git
        ]
        grep.extend([f":(exclude){glob}" for glob in whitelist])
        sed = f'sed --regexp-extended "s/(.*?):.*?({future_tag_regex}).*?/\\1:\\2/g"'
        lines = commands.get_pipe_command_output(
            grep,
            piped_to=sed,
            cwd=directory,
            valid_return_codes=(0, 1),  # 0 if there are matches, 1 otherwise
        )

    occurrences = collections.defaultdict(list)
    for line in lines:
        path, line_no, tag = line.split(":", 2)
        occurrences[tag].append(FutureTag(path=path, line_no=int(line_no), tag=tag))
    return occurrences


def get_orphan_futures(config):
    """Return orphan FUTURE tags.

    A FUTURE tag is orphan if it does not also appear (elsewhere) on a
    line with an annotation.
    """
    known_tags = get_known_future_tags(
        config.path,
        fr'{config.ignored_orphans_annotations_regex}|{config.annotation_regex}',
        config.future_tag_regex, config.whitelist,
    )
    futures = get_all_futures(config.path, config.future_tag_regex, config.whitelist)
    orphans = []
    for tag, occurrences in sorted(futures.items()):
        if tag in known_tags:
            continue
        occurrences = sorted(occurrences, key=lambda occ: (occ.path, occ.line_no))
        for occurrence in occurrences:
            last_committer_email, _last_modification = get_line_blame(
                occurrence.path, occurrence.line_no, cwd=config.path
            )
            occurrence.author = get_login_from_committer_email(last_committer_email)
            orphans.append(occurrence)
    return orphans



import os
import subprocess
import sys


try:
    import toml
except ImportError:  # pragma: no cover
    toml = None


PYPROJECT_FILENAME = "pyproject.toml"


def is_git_directory(path):
    res = subprocess.run(
        ["git", "rev-parse"],
        cwd=path,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
    )
    return res.returncode == os.EX_OK


def replace_dashes(options):
    """Recursively replace dashes by underscores in dictonary keys."""
    if not isinstance(options, dict):
        return options
    return {
        option.replace('-', '_'): replace_dashes(value)
        for option, value in options.items()
    }


def read_from_configuration_file(path, tool_name):
    if not toml:  # pragma: no cover
        sys.exit(
            'You must install with `pip install "check-oldies[toml]" '
            "to read from TOML configuration files."
        )
    try:
        conf = toml.load(path)
    except Exception as exc:  # pylint: disable=broad-except # pragma: no cover
        sys.exit(f"Error reading {path}: {exc.args[-1]}")

    if tool_name:
        try:
            options = conf["tool"][tool_name]
        except KeyError:  # pragma: no cover
            return {}
    else:
        options = conf

    return replace_dashes(options)


def get_config(
        tool_name,
        arg_parser,
        argv,
        config_class,
        config_file_default_name=PYPROJECT_FILENAME,
):
    """Get options from configuration file, possibly overriden by command
    line arguments.
    """
    args = arg_parser.parse_args(argv)
    if not args.conf and os.path.exists(config_file_default_name):
        args.conf = config_file_default_name
    config = {}
    if args.conf:
        config = read_from_configuration_file(args.conf, tool_name)
    # Override with command line arguments.
    for option, value in vars(args).items():
        if option != "conf" and value is not None:
            config[option] = value
    # Set defaults for missing settings and turn dict into an object.
    return config_class(**config)
